﻿/*
 * Copyright 2022 faddenSoft
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;

using PluginCommon;

namespace AtariSprite {
    /// <summary>
    /// Visualization generators for tile/sprite system in Atari Centipede (and possibly others).
    /// </summary>
    public class VisCentipede : MarshalByRefObject, IPlugin, IPlugin_Visualizer {
        // IPlugin
        public string Identifier {
            get { return "Atari Centipede Graphic Visualizer"; }
        }
        private IApplication mAppRef;
        private byte[] mFileData;
        //private AddressTranslate mAddrTrans;

        // Visualization identifiers; DO NOT change or projects that use them will break.
        private const string VIS_ROM_CHART = "gfx-rom-chart";
        private const string VIS_TILE_CHART = "gfx-tile-chart";
        private const string VIS_MOBJ_CHART = "gfx-mobj-chart";
        private const string VIS_TILE = "gfx-playfield-tile";
        private const string VIS_MOBJ = "gfx-motion-object";

        private const string P_OFFSET = "offset";
        private const string P_COLOR_PALETTE = "colorPalette";
        private const string P_SHOW_LABELS = "showLabels";
        private const string P_INDEX = "index";

        private const int GFX_ROM_SIZE = 4096;
        private const int TILE_WIDTH = 8;
        private const int TILE_HEIGHT = 8;
        private const int MOBJ_WIDTH = 16;
        private const int MOBJ_HEIGHT = 8;
        private const int MAX_PALETTE = 3;

        // Visualization descriptors.
        private VisDescr[] mDescriptors = new VisDescr[] {
            new VisDescr(VIS_ROM_CHART, "Full ROM as 8x8 Tiles", VisDescr.VisType.Bitmap,
                new VisParamDescr[] {
                    new VisParamDescr("File offset (hex)",
                        P_OFFSET, typeof(int), 0, 0x00ffffff, VisParamDescr.SpecialMode.Offset, 0),
                    new VisParamDescr("Color palette",
                        P_COLOR_PALETTE, typeof(int), 0, MAX_PALETTE, 0, 0),
                     new VisParamDescr("Show labels",
                        P_SHOW_LABELS, typeof(bool), 0, 0, 0, true),
                }),
            new VisDescr(VIS_TILE_CHART, "Tile Chart", VisDescr.VisType.Bitmap,
                new VisParamDescr[] {
                    new VisParamDescr("File offset (hex)",
                        P_OFFSET, typeof(int), 0, 0x00ffffff, VisParamDescr.SpecialMode.Offset, 0),
                    new VisParamDescr("Color palette",
                        P_COLOR_PALETTE, typeof(int), 0, MAX_PALETTE, 0, 0),
                     new VisParamDescr("Show labels",
                        P_SHOW_LABELS, typeof(bool), 0, 0, 0, true),
                }),
            new VisDescr(VIS_MOBJ_CHART, "Motion Object Chart", VisDescr.VisType.Bitmap,
                new VisParamDescr[] {
                    new VisParamDescr("File offset (hex)",
                        P_OFFSET, typeof(int), 0, 0x00ffffff, VisParamDescr.SpecialMode.Offset, 0),
                    new VisParamDescr("Color palette",
                        P_COLOR_PALETTE, typeof(int), 0, MAX_PALETTE, 0, 0),
                     new VisParamDescr("Show labels",
                        P_SHOW_LABELS, typeof(bool), 0, 0, 0, true),
                }),
            new VisDescr(VIS_TILE, "Playfield Tile", VisDescr.VisType.Bitmap,
                new VisParamDescr[] {
                    new VisParamDescr("File offset (hex)",
                        P_OFFSET, typeof(int), 0, 0x00ffffff, VisParamDescr.SpecialMode.Offset, 0),
                    new VisParamDescr("Index",
                        P_INDEX, typeof(int), 0, 63, 0, 0),
                    new VisParamDescr("Color palette",
                        P_COLOR_PALETTE, typeof(int), 0, MAX_PALETTE, 0, 0),
                }),
            new VisDescr(VIS_MOBJ, "Motion Object", VisDescr.VisType.Bitmap,
                new VisParamDescr[] {
                    new VisParamDescr("File offset (hex)",
                        P_OFFSET, typeof(int), 0, 0x00ffffff, VisParamDescr.SpecialMode.Offset, 0),
                    new VisParamDescr("Index",
                        P_INDEX, typeof(int), 0, 63, 0, 0),
                    new VisParamDescr("Color palette",
                        P_COLOR_PALETTE, typeof(int), 0, MAX_PALETTE, 0, 0),
                }),
        };


        // IPlugin
        public void Prepare(IApplication appRef, byte[] fileData, AddressTranslate addrTrans) {
            mAppRef = appRef;
            mFileData = fileData;
            //mAddrTrans = addrTrans;
        }

        // IPlugin
        public void Unprepare() {
            mAppRef = null;
            mFileData = null;
            //mAddrTrans = null;
        }

        // IPlugin_Visualizer
        public VisDescr[] GetVisGenDescrs() {
            if (mFileData == null) {
                return null;
            }
            return mDescriptors;
        }

        // IPlugin_Visualizer
        public IVisualization2d Generate2d(VisDescr descr,
                ReadOnlyDictionary<string, object> parms) {
            switch (descr.Ident) {
                case VIS_ROM_CHART:
                    return GenerateTileChart(parms, true);
                case VIS_TILE_CHART:
                    return GenerateTileChart(parms, false);
                case VIS_MOBJ_CHART:
                    return GenerateMobjChart(parms);
                case VIS_TILE:
                    return GenerateSingle(parms, false);
                case VIS_MOBJ:
                    return GenerateSingle(parms, true);
                default:
                    mAppRef.ReportError("Unknown ident " + descr.Ident);
                    return null;
            }
        }

        /// <summary>
        /// Generates a chart with all 64 playfield tiles, or of the entire gfx ROM as
        /// if it held 256 tiles.
        /// </summary>
        private IVisualization2d GenerateTileChart(ReadOnlyDictionary<string, object> parms,
                bool showFullRom) {
            int gfxOffset = Util.GetFromObjDict(parms, P_OFFSET, 0);
            int paletteNum = Util.GetFromObjDict(parms, P_COLOR_PALETTE, 0);
            bool showLabels = Util.GetFromObjDict(parms, P_SHOW_LABELS, true);

            if (gfxOffset < 0 || gfxOffset + GFX_ROM_SIZE > mFileData.Length) {
                mAppRef.ReportError("Invalid offset");
                return null;
            }

            const int numCols = 16;
            int numRows = 16;
            if (!showFullRom) {
                numRows = 4;
            }

            const int spacing = 1;
            const int tWidth = TILE_WIDTH + spacing;
            const int tHeight = TILE_HEIGHT + spacing;
            int leftLabelSpacing = showLabels ? 1 + 8 + 8 : 0;
            int topLabelSpacing = showLabels ? 1 + 8 : 0;

            VisBitmap8 vb = new VisBitmap8(tWidth * numCols + leftLabelSpacing + 1,
                tHeight * numRows + topLabelSpacing + 1);
            SetPalette(vb, paletteNum);

            if (showLabels) {
                // Draw hex digits on the top and sides.
                for (int i = 0; i < numCols; i++) {
                    char ch = (i < 10) ? (char)('0' + i) : (char)('A' + i - 10);
                    VisBitmap8.DrawChar(vb, ch, leftLabelSpacing + i * tWidth + 1, 1,
                        (byte)Color.TextFG, (byte)Color.TextBG);
                }
                for (int i = 0; i < numRows; i++) {
                    char ch = (i < 10) ? (char)('0' + i) : (char)('A' + i - 10);
                    VisBitmap8.DrawChar(vb, ch, 1, topLabelSpacing + i * tHeight + 1,
                        (byte)Color.TextFG, (byte)Color.TextBG);
                    VisBitmap8.DrawChar(vb, '0', 9, topLabelSpacing + i * tHeight + 1,
                        (byte)Color.TextFG, (byte)Color.TextBG);
                }
            }

            for (int idx = 0; idx < numCols * numRows; idx++) {
                int xc = (idx & 0x0f) * tWidth + leftLabelSpacing + 1;
                int yc = ((idx & 0xff) >> 4) * tHeight + topLabelSpacing + 1;

                // If we're just doing the 64-item playfield tile chart, offset the index
                // to skip over the motion object area.
                RenderTile(gfxOffset, idx + (showFullRom ? 0 : 64), vb, false, xc, yc);
            }

            return vb;
        }

        /// <summary>
        /// Generates a chart with all 64 motion objects.
        /// </summary>
        private IVisualization2d GenerateMobjChart(ReadOnlyDictionary<string, object> parms) {
            int gfxOffset = Util.GetFromObjDict(parms, P_OFFSET, 0);
            int paletteNum = Util.GetFromObjDict(parms, P_COLOR_PALETTE, 0);
            bool showLabels = Util.GetFromObjDict(parms, P_SHOW_LABELS, true);

            if (gfxOffset < 0 || gfxOffset + GFX_ROM_SIZE > mFileData.Length) {
                mAppRef.ReportError("Invalid offset");
                return null;
            }

            // The sprites are 2x as wide as playfield tiles, so it's a little nicer to show
            // the image grid as 8x8 rather than 16x4.  OTOH, requires user to do hex math.
            const int numCols = 8;
            const int numRows = 8;

            const int spacing = 1;
            const int tWidth = MOBJ_WIDTH + spacing;
            const int tHeight = MOBJ_HEIGHT + spacing;
            int leftLabelSpacing = showLabels ? 1 + 8 + 8 : 0;
            int topLabelSpacing = showLabels ? 1 + 8 : 0;

            VisBitmap8 vb = new VisBitmap8(tWidth * numCols + leftLabelSpacing + 1,
                tHeight * numRows + topLabelSpacing + 1);
            SetPalette(vb, paletteNum);

            if (showLabels) {
                // Draw hex digits on the top and sides.
                for (int i = 0; i < numCols; i++) {
                    char ch = (i < 10) ? (char)('0' + i) : (char)('A' + i - 10);
                    VisBitmap8.DrawChar(vb, ch, leftLabelSpacing + i * tWidth + 1 + 4, 1,
                        (byte)Color.TextFG, (byte)Color.TextBG);
                }
                for (int i = 0; i < numRows; i++) {
                    int label = i * numCols;
                    char ch1 = (char)('0' + (label >> 4));
                    char ch2 = (char)('0' + (label & 0x0f));
                    VisBitmap8.DrawChar(vb, ch1, 1, topLabelSpacing + i * tHeight + 1,
                        (byte)Color.TextFG, (byte)Color.TextBG);
                    VisBitmap8.DrawChar(vb, ch2, 9, topLabelSpacing + i * tHeight + 1,
                        (byte)Color.TextFG, (byte)Color.TextBG);
                }
            }

            for (int idx = 0; idx < numCols * numRows; idx++) {
                int xc = (idx & 0x07) * tWidth + leftLabelSpacing + 1;
                int yc = ((idx & 0xff) >> 3) * tHeight + topLabelSpacing + 1;

                RenderTile(gfxOffset, idx, vb, true, xc, yc);
            }

            return vb;
        }

        /// <summary>
        /// Generates a bitmap with a single playfield tile or motion object graphic.
        /// </summary>
        private IVisualization2d GenerateSingle(ReadOnlyDictionary<string, object> parms,
                bool isMobj) {
            int gfxOffset = Util.GetFromObjDict(parms, P_OFFSET, 0);
            int index = Util.GetFromObjDict(parms, P_INDEX, 0);
            int paletteNum = Util.GetFromObjDict(parms, P_COLOR_PALETTE, 0);

            if (gfxOffset < 0 || gfxOffset + GFX_ROM_SIZE > mFileData.Length) {
                mAppRef.ReportError("Invalid offset");
                return null;
            }
            if (index < 0 || index > 63) {
                mAppRef.ReportError("Invalid index");
                return null;
            }
            if (!isMobj) {
                index += 64;
            }

            VisBitmap8 vb = new VisBitmap8(isMobj ? 16 : 8, 8);
            SetPalette(vb, paletteNum);
            RenderTile(gfxOffset, index, vb, isMobj, 0, 0);
            return vb;
        }

        /// <summary>
        /// Renders a tile from the GFX ROM.
        /// </summary>
        /// <remarks>
        /// Tiles are rotated 90 degrees from how bitmaps are normally stored.  The first byte
        /// is the leftmost column rather than the top row.
        /// </remarks>
        /// <param name="gfxOffset">Offset of gfx ROM.</param>
        /// <param name="itemIndex">Item index (0-256 for tiles, 0-63 for motion objs).</param>
        /// <param name="vb">Bitmap to render to.</param>
        /// <param name="isMobj">True if this is a motion object.</param>
        /// <param name="xc">X coordinate for upper-left coordinate.</param>
        /// <param name="yc">Y coordinate for upper-left coordinate.</param>
        private void RenderTile(int gfxOffset, int itemIndex, VisBitmap8 vb, bool isMobj,
                int xc, int yc) {
            const int PLANE_OFFSET = 2048;  // offset to second bit plane

            int tileOff;
            if (isMobj) {
                // There are 64 motion objects.  The even-numbered objects start at +0,
                // the odd objects start at +1024.  Adjust the index.
                itemIndex = ((itemIndex & 0x3e) >> 1) | ((itemIndex & 0x01) << 6);
                tileOff = gfxOffset + itemIndex * 16;
            } else {
                tileOff = gfxOffset + itemIndex * 8;
            }
            for (int col = 0; col < (isMobj ? 16 : 8); col++) {
                byte part0 = mFileData[tileOff];
                byte part1 = mFileData[tileOff + PLANE_OFFSET];
                for (int bit = 7; bit >= 0; bit--) {
                    int val = ((part0 >> bit) & 0x01) | (((part1 >> bit) & 0x01) << 1);
                    vb.SetPixelIndex(xc, yc + bit,
                        (byte)((byte)Color.Color0 + val));
                }

                tileOff++;
                xc++;
            }
        }

        private enum Color : byte {
            Transparent = 0,
            TextFG = 1,
            TextBG = 2,
            Color0 = 3,
            Color1 = 4,
            Color2 = 5,
            Color3 = 6
        }

        private void SetPalette(VisBitmap8 vb, int pal) {
            vb.SetColor((byte)Color.Transparent, 0x00, 0x00, 0x00, 0x00);
            vb.SetColor((byte)Color.TextFG, 0xff, 0x00, 0x00, 0x00);            // black
            vb.SetColor((byte)Color.TextBG, 0xff, 0xff, 0xff, 0xff);            // white

            // Currently have three palettes:
            //  #0: wave 1 tile colors
            //  #1: wave 1 motion object colors (colors 1 and 3 are swapped)
            //  #2: wave 1 motion object colors for centipede body (color #2 == color #3)
            //  #2: wave 2 motion object tile colors with transparent background
            switch (pal) {
                default:
                case 0:
                    vb.SetColor((byte)Color.Color0, 0xff, 0x00, 0x00, 0x00);    // black
                    vb.SetColor((byte)Color.Color1, 0xff, 0x00, 0xff, 0x00);    // green
                    vb.SetColor((byte)Color.Color2, 0xff, 0xff, 0x00, 0x00);    // red
                    vb.SetColor((byte)Color.Color3, 0xff, 0xff, 0xff, 0xc0);    // beige
                    break;
                case 1:
                    vb.SetColor((byte)Color.Color0, 0xff, 0x00, 0x00, 0x00);    // black
                    vb.SetColor((byte)Color.Color1, 0xff, 0xff, 0xff, 0xc0);    // beige
                    vb.SetColor((byte)Color.Color2, 0xff, 0xff, 0x00, 0x00);    // red
                    vb.SetColor((byte)Color.Color3, 0xff, 0x00, 0xff, 0x00);    // green
                    break;
                case 2:
                    vb.SetColor((byte)Color.Color0, 0xff, 0x00, 0x00, 0x00);    // black
                    vb.SetColor((byte)Color.Color1, 0xff, 0xff, 0xff, 0xc0);    // beige
                    vb.SetColor((byte)Color.Color2, 0xff, 0x00, 0xff, 0x00);    // green
                    vb.SetColor((byte)Color.Color3, 0xff, 0x00, 0xff, 0x00);    // green
                    break;
                case 3:
                    vb.SetColor((byte)Color.Color0, 0x00, 0x00, 0x00, 0x00);    // transparent
                    vb.SetColor((byte)Color.Color1, 0xff, 0xff, 0xc0, 0x00);    // orange
                    vb.SetColor((byte)Color.Color2, 0xff, 0x00, 0xff, 0xc0);    // g-cyan
                    vb.SetColor((byte)Color.Color3, 0xff, 0xff, 0x00, 0xc0);    // r-magenta
                    break;
            }
        }
    }
}
